/* ndk - [ console.c ]
 *
 * Contains the minimal ammount of code necessary
 * to have console output for the kernel.
 *
 * (c)2002 dcipher / neuraldk
 *           www.neuraldk.org
 */

#include "console.h"

/* The video memory address. */
#define COLOUR_VIDEO    0xB8000
#define MONO_VIDEO	0xB0000

/* The number of lines and columns */
long consoleCols   = 80;
long consoleLines  = 24;
/* precalculate some commonly used equations */
long consoleBytesPerLine = 80 * 2;
long consoleTotalBytes   = 4000;
/* The attribute of an character. */
char consoleAttrib = 15;

/* position of cursor */
static int consoleXPos, consoleYPos;

/* location of video memory. */
static volatile unsigned char *consoleAddr;

//void (*printf)(const char *format, ...) = kprintf;

/* Clear the screen and initialize VIDEO, XPOS and YPOS. */
void cls (void) {
  int i;

  // check for mono adaptor?
  consoleAddr = (unsigned char *) COLOUR_VIDEO;

  for (i = 0; i < consoleCols * consoleLines * 2; i++)
    *(consoleAddr + i) = 0;

  consoleXPos = 0;
  consoleYPos = 0;
}

/* Convert the integer D to a string and save the string in BUF. If
   BASE is equal to 'd', interpret that D is decimal, and if BASE is
   equal to 'x', interpret that D is hexadecimal. */
static void itoa (char *buf, int base, int d, signed int length) {
  char *p = buf;
  char *p1, *p2;
  unsigned long ud = d;
  int divisor = 10;

  /* If %d is specified and D is minus, put `-' in the head. */
  if (base == 'd' && d < 0) {
    *p++ = '-';
    buf++;
    ud = -d;
  } else if (base == 'x')
    divisor = 16;

  /* Divide UD by DIVISOR until UD == 0. */
  do {
    int remainder = ud % divisor;

    *p++ = (remainder < 10) ? remainder + '0' : remainder + 'a' - 10;
    length--;
  } while (ud /= divisor);

  while(length > 0) {
    *p++ = '0';
    length--;
  }

  /* Terminate BUF. */
  *p = 0;

  /* Reverse BUF. */
  p1 = buf;
  p2 = p - 1;
  while (p1 < p2) {
    char tmp = *p1;
    *p1 = *p2;
    *p2 = tmp;
    p1++;
    p2--;
  }
}

/* Put the character C on the screen. */
static void putchar (int c) {
  if (c == '\n' || c == '\r') {
  newline:
    consoleXPos = 0;
    consoleYPos++;
    if(consoleYPos >= consoleLines) {
      consoleYPos = consoleLines;
      // scroll up a line
      memcpy(consoleAddr,
             consoleAddr + consoleBytesPerLine,
             consoleTotalBytes - consoleBytesPerLine );
      // clear the last line
      memset(consoleAddr + consoleTotalBytes - consoleBytesPerLine,
             0,
             consoleBytesPerLine);
    }
    return;
  } else if (c == '\t') {
    consoleXPos += (8 - (consoleXPos & 7));
    return;
  }

  *(consoleAddr + (consoleXPos + consoleYPos * consoleCols) * 2)     = c & 0xFF;
  *(consoleAddr + (consoleXPos + consoleYPos * consoleCols) * 2 + 1) = consoleAttrib;

  consoleXPos++;
  if (consoleXPos >= consoleCols)
    goto newline;
}

/* Format a string and print it on the screen, just like the libc
   function printf. */
void printf (const char *format, ...) {
  char **arg = (char **) &format;
  int c;
  char buf[20];
  int length;

  arg++;

  while ((c = *format++) != 0) {
    length = 0;
    if (c != '%')
      putchar (c);
    else {
      char *p;

      c = *format++;
      while(c >= '0' && c <= '9') {
        length *= 10;
	length += c - '0';
	c = *format++;
      }
      switch (c) {
        case 'd':
        case 'u':
        case 'x':
          itoa (buf, c, *((int *) arg++), length);
          p = buf;
          goto string;
          break;

        case 's':
          p = *arg++;
          if (! p)
            p = "(null)";

        string:
          while (*p)
          putchar (*p++);
          break;

        default:
          putchar (*((int *) arg++));
          break;
      }
    }
  }
}