/* ndk - [ timer.c ]
 *
 * Contains code to initialize and control
 * the x86's programmable interval timer.
 *
 * (c)2002 dcipher / neuraldk
 *           www.neuraldk.org
 */

#include "timer.h"
#include "array.h" // just a test

long timerTicks = 0;

void timerInit(void) {
  printf("timerInit: ");
  createGate(&IDT, 0x20, (long)int20h_wrapper, SEL_P0_CODE, D_TRAP);
  timerSetFreq(100);  // 100 times per second
  picEnableIRQ(0);
  printf("pass\n");
  printf("enabling ints: ");
  asm("sti");
  printf("pass\n");
}

void timerSetFreq(short freq) {
  long counter = PIT_FREQ / freq;

  outportb(TIMER_CTRL, /*0x43*/TIMER0 + TIMER_WHOLE + TIMER_MODE2);
  outportb(TIMER0_CNTR, counter % 256);  // send LSB
  outportb(TIMER0_CNTR, counter / 256);  // send MSB
  return;
}

void timerHandler() {
  char *scr = (char *)0x000b8000;

  // VIDINC scr[0]++;
  timerTicks++;

  outportb (M_PIC, EOI);
  timerScheduler();
}

void timerScheduler(void) {
  char *desc;

  /* if there's 0 or 1 task, no need to cycle */
  if(firstTask == NULL ||
     firstTask->next == NULL) return;

  /* standard round robin scheduler */
  do {
    currentTask = (Task *)currentTask->next;
    if(currentTask == NULL)
      currentTask = (Task *)firstTask;
  } while(taskNotBusy(currentTask));

  /* check to see if current task's busy bit is set (just curious... */
  /*
  desc = (char *)arrayElementAt(&GDT, currentTask->selector / 8);
  desc+=5;
  if(*desc & 2) printf("BUSY!\n");
  else printf("N/B\n");
   */

  //printf("jmp to %x\n", currentTask->selector);
  // this is just a test... regarding my taskGate!!!
     asm("movl $0xffffffff, %%eax\n"
        "addl $0x1, %%eax\n"
        :
        :
        : "eax");
  jumpToTSS(currentTask->selector);
}

long timerGetTicks(void) {
  return timerTicks;
}
